@@ -1,7 +1,7 @@
 (********************************************************************)
 (*                                                                  *)
 (*  The Why3 Verification Platform   /   The Why3 Development Team  *)
-(*  Copyright 2010-2023 --  Inria - CNRS - Paris-Saclay University  *)
+(*  Copyright 2010-2024 --  Inria - CNRS - Paris-Saclay University  *)
 (*                                                                  *)
 (*  This software is distributed under the terms of the GNU Lesser  *)
 (*  General Public License version 2.1, with the special exception  *)
@@ -10,10 +10,10 @@
 (********************************************************************)
 
 open Mysexplib.Std_big_int [@@warning "-33"]
-open Big_int
+open Big_int_Z
 
 type t = big_int
-[@@deriving sexp_of]
+[@@deriving sexp]
 let compare = compare_big_int
 
 let zero = zero_big_int
